#define STRICT
#include "nvafx.h"
#include "HardwareShadowMap.h"
#include "HardwareShadowMapApp.h"

#define DEPTH_TEXT "Depth Bias:  %.6f"
#define SLOPE_TEXT "Slope Scale Bias:  %.2f"
#define DEPTH_ALG g_pHardwareShadowMap->m_fDepthBias = ((float)g_HUD.GetSlider(IDC_DEPTH)->GetValue())/1000000.0f
#define SLOPE_ALG g_pHardwareShadowMap->m_fBiasSlope = ((float)g_HUD.GetSlider(IDC_SLOPE)->GetValue())/100.0f

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, int )
{
    // Set the callback functions. These functions allow the sample framework to notify
    // the application about device changes, user input, and windows messages.  The 
    // callbacks are optional so you need only set callbacks for events you're interested 
    // in. However, if you don't handle the device reset/lost callbacks then the sample 
    // framework won't be able to reset your device since the application must first 
    // release all device resources before resetting.  Likewise, if you don't handle the 
    // device created/destroyed callbacks then the sample framework won't be able to 
    // recreate your device resources.
    DXUTSetCallbackDeviceCreated( OnCreateDevice );
    DXUTSetCallbackDeviceReset( OnResetDevice );
    DXUTSetCallbackDeviceLost( OnLostDevice );
    DXUTSetCallbackDeviceDestroyed( OnDestroyDevice );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( KeyboardProc );
    DXUTSetCallbackFrameRender( OnFrameRender );
    DXUTSetCallbackFrameMove( OnFrameMove );

    // Show the cursor and clip it when in full screen
    DXUTSetCursorSettings( true, true );

    InitApp();

    // Initialize the sample framework and create the desired Win32 window and Direct3D 
    // device for the application. Calling each of these functions is optional, but they
    // allow you to set several options which control the behavior of the framework.
    DXUTInit( true, true, true ); // Parse the command line, handle the default hotkeys, and show msgboxes
    DXUTCreateWindow( L"HardwareShadowMap 1.0 (DX9) (HLSL)" );
    DXUTCreateDevice( D3DADAPTER_DEFAULT, true, 640, 576, IsDeviceAcceptable, ModifyDeviceSettings );

    // Pass control to the sample framework for handling the message pump and 
    // dispatching render calls. The sample framework will call your FrameMove 
    // and FrameRender callback when there is idle time between handling window messages.
    DXUTMainLoop();

    // Perform any application-level cleanup here. Direct3D device resources are released within the
    // appropriate callback functions and therefore don't require any cleanup code here.

    return DXUTGetExitCode();
}


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
	g_pHardwareShadowMap = new HardwareShadowMap();

	// Initialize dialogs
    g_SettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );
    g_HUD.SetCallback( OnGUIEvent );
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, 34, 125, 22 );
    g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, 58, 125, 22, VK_F3 );
    g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, 82, 125, 22, VK_F2 );
    g_HUD.AddButton(IDC_LAME, L"Toggle Filtering", 35, 106,125, 22);

	g_HUD.AddSlider(IDC_DEPTH,0,0,150,22,0,400,(int)(g_pHardwareShadowMap->m_fDepthBias*1000000.0f));
	g_HUD.AddSlider(IDC_SLOPE,0,0,150,22,0,500,(int)(g_pHardwareShadowMap->m_fBiasSlope*100.0f));

	TCHAR sz[100];

	_sntprintf( sz, 100, TEXT(DEPTH_TEXT), DEPTH_ALG); sz[99] = 0;
	g_HUD.AddStatic(IDC_DEPTH_STATIC,sz,0,0,140,22);
	_sntprintf( sz, 100, TEXT(SLOPE_TEXT), SLOPE_ALG); sz[99] = 0;
	g_HUD.AddStatic(IDC_SLOPE_STATIC,sz,0,0,140,22);

	g_HUD.AddButton(IDC_WIREFRM,_T("(W)ireframe"),0,0,125,22,L'W');
	g_HUD.AddButton(IDC_PAUSE,_T("(P)ause"),0,0,125,22,L'P');


}


//--------------------------------------------------------------------------------------
// Called during device initialization, this code checks the device for some 
// minimum set of capabilities, and rejects those that don't pass by returning false.
//--------------------------------------------------------------------------------------
bool CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, 
                                  D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    static int nErrors = 0;     // use this to only show the very first error messagebox
    int nPrevErrors = nErrors;

    // check vertex shading support
    if(pCaps->VertexShaderVersion < D3DVS_VERSION(1,1))
        if (!nErrors++) 
            MessageBox(NULL, _T("Device does not support 1.1 vertex shaders!"), _T("ERROR"), MB_OK|MB_SETFOREGROUND|MB_TOPMOST);

    // check pixel shader support
    if(pCaps->PixelShaderVersion < D3DPS_VERSION(1,1))
        if (!nErrors++) 
            MessageBox(NULL, _T("Device does not support 1.1 vertex shaders!"), _T("ERROR"), MB_OK|MB_SETFOREGROUND|MB_TOPMOST);

    if (!(pCaps->TextureCaps & D3DPTEXTURECAPS_MIPMAP))
        if (!nErrors++) 
            MessageBox(NULL, _T("Device does not support mipmaps!"), _T("ERROR"), MB_OK|MB_SETFOREGROUND|MB_TOPMOST);
    
    if(pCaps->MaxSimultaneousTextures < 2)
        if (!nErrors++) 
            MessageBox(NULL, _T("Device does not support two simultaneous textures!"), _T("ERROR"), MB_OK|MB_SETFOREGROUND|MB_TOPMOST);

    if(!(pCaps->RasterCaps & D3DPRASTERCAPS_SLOPESCALEDEPTHBIAS))
        if (!nErrors++) 
            MessageBox(NULL, _T("Device does not support slope scale depth bias!"), _T("ERROR"), MB_OK|MB_SETFOREGROUND|MB_TOPMOST);

    return (nErrors > nPrevErrors) ? false : true;
}


//--------------------------------------------------------------------------------------
// This callback function is called immediately before a device is created to allow the 
// application to modify the device settings. The supplied pDeviceSettings parameter 
// contains the settings that the framework has selected for the new device, and the 
// application can make any desired changes directly to this structure.  Note however that 
// the sample framework will not correct invalid device settings so care must be taken 
// to return valid device settings, otherwise IDirect3D9::CreateDevice() will fail.  
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, const D3DCAPS9* pCaps, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been 
// created, which will happen during application initialization and windowed/full screen 
// toggles. This is the best location to create D3DPOOL_MANAGED resources since these 
// resources need to be reloaded whenever the device is destroyed. Resources created  
// here should be released in the OnDestroyDevice callback. 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    V_RETURN( g_DialogResourceManager.OnCreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnCreateDevice( pd3dDevice ) );

    // Initialize the font
    V_RETURN( D3DXCreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, 
                         OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
                         L"Arial", &g_pFont ) );

    // Set up our view matrix. A view matrix can be defined given an eye point and
    // a point to lookat. Here, we set the eye five units back along the z-axis and 
	// up three units and look at the origin.
	
    D3DXVECTOR3 vFromPt   = D3DXVECTOR3(100.0f, 100.0f, -80.0f);
    D3DXVECTOR3 vLookatPt = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	g_Camera.SetViewParams( &vFromPt, &vLookatPt);
	
	
	/*
	D3DXMATRIX *m_View = g_Camera.GetViewMatrix();
	//using pre-generated view matrix for nice starting position
    m_View->_11 =  0.5594f; m_View->_12 = -0.3993f; m_View->_13 =  -0.7263f; m_View->_14 = 0.0f;
    m_View->_21 = -0.0404f; m_View->_22 =  0.8621f; m_View->_23 =  -0.5051f; m_View->_24 = 0.0f;
    m_View->_31 =  0.8279f; m_View->_32 =  0.3119f; m_View->_33 =   0.4662f; m_View->_34 = 0.0f;
    m_View->_41 =  5.4500f; m_View->_42 = -7.9000f; m_View->_43 = 123.7043f; m_View->_44 = 1.0f;

	D3DXMATRIX *view = m_View;
    view->_41 = view->_42 = view->_43 = 0.0f;
	*/
	return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been 
// reset, which will happen after a lost device scenario. This is the best location to 
// create D3DPOOL_DEFAULT resources since these resources need to be reloaded whenever 
// the device is lost. Resources created here should be released in the OnLostDevice 
// callback. 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, 
                                const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	HRESULT hr;

    V_RETURN( g_DialogResourceManager.OnResetDevice() );
    V_RETURN( g_SettingsDlg.OnResetDevice() );

    if( g_pFont )
        V_RETURN( g_pFont->OnResetDevice() );

    // Create a sprite to help batch calls when drawing many lines of text
    V_RETURN( D3DXCreateSprite( pd3dDevice, &g_pTextSprite ) );

    // Setup the camera's projection parameters
    g_Camera.SetProjParams( D3DXToRadian(60.0f), 1.0f, 10.0f, 800.0f );
    g_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );

    g_HUD.SetLocation( 0, 0 );
    g_HUD.SetSize( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );

	int iY = 10;
    g_HUD.GetControl( IDC_TOGGLEFULLSCREEN )->SetLocation( pBackBufferSurfaceDesc->Width - 135, iY);
    g_HUD.GetControl( IDC_TOGGLEREF )->SetLocation( pBackBufferSurfaceDesc->Width - 135, iY += 24 );
    g_HUD.GetControl( IDC_CHANGEDEVICE )->SetLocation( pBackBufferSurfaceDesc->Width - 135, iY += 24 );
	g_HUD.GetControl(IDC_WIREFRM)->SetLocation( pBackBufferSurfaceDesc->Width - 135, iY += 24);
	g_HUD.GetControl(IDC_PAUSE)->SetLocation( pBackBufferSurfaceDesc->Width - 135, iY += 24);
    g_HUD.GetControl(IDC_LAME)->SetLocation( pBackBufferSurfaceDesc->Width - 135, iY += 24);

	#define C_WIDTH 170
	iY = 30;
	g_HUD.GetControl(IDC_SLOPE)->SetLocation( pBackBufferSurfaceDesc->Width - C_WIDTH, pBackBufferSurfaceDesc->Height - (iY));
	g_HUD.GetControl(IDC_SLOPE_STATIC)->SetLocation( pBackBufferSurfaceDesc->Width - C_WIDTH, pBackBufferSurfaceDesc->Height - (iY += 22));
	g_HUD.GetControl(IDC_DEPTH)->SetLocation( pBackBufferSurfaceDesc->Width - C_WIDTH, pBackBufferSurfaceDesc->Height - (iY += 22));
	g_HUD.GetControl(IDC_DEPTH_STATIC)->SetLocation( pBackBufferSurfaceDesc->Width - C_WIDTH, pBackBufferSurfaceDesc->Height - (iY += 22));


    pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

	g_pHardwareShadowMap->m_Projection = *g_Camera.GetProjMatrix();
	V_RETURN(g_pHardwareShadowMap->ResetDevice( pd3dDevice, pBackBufferSurfaceDesc ));

    return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called once at the beginning of every frame. This is the
// best location for your application to handle updates to the scene, but is not 
// intended to contain actual rendering calls, which should instead be placed in the 
// OnFrameRender callback.  
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    // TODO: update world
    g_Camera.FrameMove( fElapsedTime );
}


//--------------------------------------------------------------------------------------
// This callback function will be called at the end of every frame to perform all the 
// rendering calls for the scene, and it will also be called if the window needs to be 
// repainted. After this function has returned, the sample framework will call 
// IDirect3DDevice9::Present to display the contents of the next buffer in the swap chain
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    // If the settings dialog is being shown, then
    // render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }

    HRESULT hr;

    // Clear the viewport
    pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                        0x000000ff, 1.0f, 0L);


	// Begin the scene
    if (SUCCEEDED(pd3dDevice->BeginScene()))
    {
		g_pHardwareShadowMap->Render(pd3dDevice, fTime, fElapsedTime, 
			g_Camera.GetWorldMatrix(), 
			g_Camera.GetViewMatrix(), 
			g_Camera.GetProjMatrix());

        // Render stats and help text  
        RenderText();

		V( g_HUD.OnRender( fElapsedTime ) );

        // End the scene.
        V( pd3dDevice->EndScene() );
    }
}


//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
//--------------------------------------------------------------------------------------
void RenderText()
{
    // The helper object simply helps keep track of text position, and color
    // and then it calls pFont->DrawText( m_pSprite, strMsg, -1, &rc, DT_NOCLIP, m_clr );
    // If NULL is passed in as the sprite object, then it will work however the 
    // pFont->DrawText() will not be batched together.  Batching calls will improves performance.
    CDXUTTextHelper txtHelper( g_pFont, g_pTextSprite, 15 );

    // Output statistics
    txtHelper.Begin();
    txtHelper.SetInsertionPos( 5, 5 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
    txtHelper.DrawTextLine( DXUTGetFrameStats() );
    txtHelper.DrawTextLine( DXUTGetDeviceStats() );

    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
    
    // Draw help
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
    txtHelper.DrawTextLine( L"Press ESC to quit" );
	txtHelper.DrawTextLine( L"Left mouse to move model");
	txtHelper.DrawTextLine( L"Right mouse to move scene");
	txtHelper.DrawTextLine( L"Mouse wheel to zoom");

    const WCHAR* techs[4] = { L"UseHardwareShadowMap", L"UseHardwareShadowMapLame", L"UseHardwareShadowMapGoodNoRot", L"UseHardwareShadowMapGoodRot" };
    txtHelper.DrawTextLine(techs[g_pHardwareShadowMap->m_bLameTech]);

    txtHelper.End();
}


//--------------------------------------------------------------------------------------
// Before handling window messages, the sample framework passes incoming windows 
// messages to the application through this callback function. If the application sets 
// *pbNoFurtherProcessing to TRUE, then the sample framework will not process this message.
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
    // Always allow dialog resource manager calls to handle global messages
    // so GUI state is updated correctly
    g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );

    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass all remaining windows messages to camera so it can respond to user input
    g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}


//--------------------------------------------------------------------------------------
// As a convenience, the sample framework inspects the incoming windows messages for
// keystroke messages and decodes the message parameters to pass relevant keyboard
// messages to the application.  The framework does not remove the underlying keystroke 
// messages, which are still passed to the application's MsgProc callback.
//--------------------------------------------------------------------------------------
void CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
}


//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
	switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN: DXUTToggleFullScreen(); break;
        case IDC_TOGGLEREF:        DXUTToggleREF(); break;
        case IDC_CHANGEDEVICE:     g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); break;
		case IDC_DEPTH:
			{
				TCHAR sz[100];
				_sntprintf( sz, 100, TEXT(DEPTH_TEXT), DEPTH_ALG); sz[99] = 0;
				g_HUD.GetStatic(IDC_DEPTH_STATIC)->SetText(sz);
				break;
			}
		case IDC_SLOPE:
			{
				TCHAR sz[100];
				_sntprintf( sz, 100, TEXT(SLOPE_TEXT), SLOPE_ALG); sz[99] = 0;
				g_HUD.GetStatic(IDC_SLOPE_STATIC)->SetText(sz);
				break;
			}
		case IDC_WIREFRM:
			{
				if(g_pHardwareShadowMap->m_bWireframe = !g_pHardwareShadowMap->m_bWireframe)
					g_HUD.GetButton(IDC_WIREFRM)->SetText(_T("Un(W)ireframe"));
				else
					g_HUD.GetButton(IDC_WIREFRM)->SetText(_T("(W)ireframe"));
				break;
				break;
			}
		case IDC_PAUSE:
			{
				if(g_pHardwareShadowMap->m_Paused = !g_pHardwareShadowMap->m_Paused)
					g_HUD.GetButton(IDC_PAUSE)->SetText(_T("Un(P)ause"));
				else
					g_HUD.GetButton(IDC_PAUSE)->SetText(_T("(P)ause"));
				break;
			}
        case IDC_LAME:
            {
                g_pHardwareShadowMap->m_bLameTech = (g_pHardwareShadowMap->m_bLameTech+1) % 3;
                break;
            }

    }
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has 
// entered a lost state and before IDirect3DDevice9::Reset is called. Resources created
// in the OnResetDevice callback should be released here, which generally includes all 
// D3DPOOL_DEFAULT resources. See the "Lost Devices" section of the documentation for 
// information about lost devices.
//--------------------------------------------------------------------------------------
void CALLBACK OnLostDevice( void* pUserContext )
{
    g_DialogResourceManager.OnLostDevice();
    g_SettingsDlg.OnLostDevice();

    if( g_pFont )
        g_pFont->OnLostDevice();

	SAFE_RELEASE(g_pTextSprite);
	g_pHardwareShadowMap->LostDevice();
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has 
// been destroyed, which generally happens as a result of application termination or 
// windowed/full screen toggles. Resources created in the OnCreateDevice callback 
// should be released here, which generally includes all D3DPOOL_MANAGED resources. 
//--------------------------------------------------------------------------------------
void CALLBACK OnDestroyDevice( void* pUserContext )
{
    g_DialogResourceManager.OnDestroyDevice();
    g_SettingsDlg.OnDestroyDevice();

    SAFE_RELEASE(g_pFont);
}